/*
 * eth_ecat_sdo_mailbox.hpp
 *
 *  Created on: May 16, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_SDO_MAILBOX_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_SDO_MAILBOX_HPP_

#include "ethernet_industry/eth_ecat_buffer.hpp"

namespace free_rtos {

namespace ecat_sdo_mailbox {

enum class SDOReqCommandSpecifier : std::uint16_t {
    DOWNLOAD_SEGMENT = 0x0,
    DOWNLOAD,
    UPLOAD,
    UPLOAD_SEGMENT,
    ABORT_TRANSFER
};

enum class TransferType : std::uint16_t {
    NORMAL = 0x0,
    EXPEDITED
};

enum class SDORespCommandSpecifier : std::uint16_t {
    UPLOAD_SEGMENT = 0x0,
    DOWNLOAD_SEGMENT,
    UPLOAD,
    DOWNLOAD,
    ABORT_TRANSFER
};

enum class Service : uint16_t {
    EMERGENCY = 0x1,
    SDO_REQUEST,
    SDO_RESPONSE,
    TX_PDO,
    RX_PDO,
    TX_PDO_REMOTE_REQUEST,
    RX_PDO_REMOTE_REQUEST,
    SDO_INFORMATION
};

struct PDOMap {
    uint16_t rx_data_size;
    uint16_t tx_data_size;
};

struct PDODescriptor {
    uint8_t size;
    uint8_t subindex;
    uint16_t index;
} __attribute__ ((packed));

struct CompleteSize {
    uint32_t value;
} __attribute__ ((packed));

struct CommandProp {
    std::uint8_t size : 1;
    std::uint8_t transfer_type : 1;
    std::uint8_t data_set_size : 2;
    std::uint8_t complete_access : 1;
    std::uint8_t command_spec : 3;
} __attribute__ ((packed));

struct CoEHeader {
    std::uint8_t number;
    std::uint8_t reserved : 4;
    std::uint8_t service : 4;
} __attribute__ ((packed));

struct CoEElements {
    CoEHeader coe_header;
    CommandProp command_specifier;
    std::uint16_t index;
    std::uint8_t subindex;
} __attribute__ ((packed));

enum class ProtocolType : std::uint16_t {
    /** Beckhoff: AoE ADS over EtherCAT */
    VENDOR_SPECIFIC = 1,
    /** Ethernet over EtherCAT */
    EoE,
    /** CanOpen Over EtherCAT */
    CoE,
    /** File access over EtherCAT */
    FoE,
    /** Servo profile over EtherCAT */
    SoE,
    /** Vendor specific (VoE) */
    VoE = 0xF
};

struct MailboxHeader {
    uint16_t length;
    uint16_t address;
    uint16_t channel : 6;
    uint16_t priority : 2;
    ProtocolType type : 4;
    uint16_t cnt : 3;
    uint16_t reserved : 1;
} __attribute__ ((packed));

} // namespace ecat_sdo_mailbox

// Специализация шаблона для распаковки протокола CoE
template<typename... TailT>
struct CustomTuple<ecat_sdo_mailbox::CoEElements, ecat_sdo_mailbox::CompleteSize, TailT...> : CustomTuple<TailT...> {
    CustomTuple(ecat_sdo_mailbox::CoEElements& head, ecat_sdo_mailbox::CompleteSize& complete_size, TailT&... tail)
    : CustomTuple<TailT...>(tail...)
    , head_(head)
    , complete_size_(complete_size) { }

    using TBase = CustomTuple<TailT...>;

    constexpr static size_t size = sizeof(ecat_sdo_mailbox::CoEElements) + sizeof(ecat_sdo_mailbox::CompleteSize) + TBase::size;

    TBase& base_ = static_cast<TBase&>(*this);
    ecat_sdo_mailbox::CoEElements& head_;
    ecat_sdo_mailbox::CompleteSize& complete_size_;

    uint8_t* pack_complete_size(uint8_t *raw) {
        if(head_.command_specifier.transfer_type == static_cast<uint8_t>(ecat_sdo_mailbox::TransferType::NORMAL)) {
            ecat_sdo_mailbox::CompleteSize* complete_size = new(raw) ecat_sdo_mailbox::CompleteSize{complete_size_};

            (void)complete_size;

            raw += sizeof(ecat_sdo_mailbox::CompleteSize);
            raw = base_.pack(raw);
        }

        if(head_.command_specifier.transfer_type == static_cast<uint8_t>(ecat_sdo_mailbox::TransferType::EXPEDITED)) {
            // По-нормальному мы не должны сюда попадать, т.к. в expedited транзакции не должно быть блока CompleteSize,
            // значит либо пользователь ошибся, либо после предыдущей отправки датаграммы слейв указал тип expedited
            raw = base_.pack(raw);
            raw += sizeof(ecat_sdo_mailbox::CompleteSize);
        }

        return raw;
    }

    uint8_t* pack(uint8_t *raw) {
        ecat_sdo_mailbox::CoEElements *head = new(raw) ecat_sdo_mailbox::CoEElements{head_};

        (void)head;

        return pack_complete_size(raw + sizeof(ecat_sdo_mailbox::CoEElements));
    }

    uint8_t* unpack_complete_size(uint8_t *raw) {
        if(head_.command_specifier.transfer_type == static_cast<uint8_t>(ecat_sdo_mailbox::TransferType::NORMAL)) {
            ecat_sdo_mailbox::CompleteSize *complete_size = reinterpret_cast<ecat_sdo_mailbox::CompleteSize*>(raw);

            complete_size_ = *complete_size;

            raw += sizeof(ecat_sdo_mailbox::CompleteSize);
            raw = base_.unpack(raw);
        }

        if(head_.command_specifier.transfer_type == static_cast<uint8_t>(ecat_sdo_mailbox::TransferType::EXPEDITED)) {
            raw = base_.unpack(raw);
            raw += sizeof(ecat_sdo_mailbox::CompleteSize);
        }

        return raw;
    }

    uint8_t* unpack(uint8_t *raw) {
        ecat_sdo_mailbox::CoEElements *head = reinterpret_cast<ecat_sdo_mailbox::CoEElements*>(raw);
        head_ = *head;

        return unpack_complete_size(raw + sizeof(ecat_sdo_mailbox::CoEElements));
    }
};

namespace ecat_sdo_mailbox {

class EcatSdoMailboxSlave {
public:
    EcatSdoMailboxSlave(ecat_buffer::EcatBufferSlave& mailbox_slave)
    : buffer_slave_(mailbox_slave) { }

    template<typename TypeT>
    void wait_available(datagram::EcatTelegram& telegram) {
        using TCommand = command::EcatCommand<TypeT, command::RD>;

        auto slave_address = buffer_slave_.get_slave().get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        uint16_t sm_status{0x0000};
        datagram::EcatDatagram<TCommand, uint16_t> datagram{ {{slave_address, buffer_regs[MailboxesRegs::AVAILABLE]}}, sm_status };

        do {
            telegram.transfer(datagram);
        } while((datagram.get_wkc() < 0x0001) || ((sm_status & 0x08) == 0));
    }

    template<typename TypeT>
    void wait_empty(datagram::EcatTelegram& telegram) {
        using TCommand = command::EcatCommand<TypeT, command::RD>;

        auto slave_address = buffer_slave_.get_slave().get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        uint8_t sm_status{0x00};
        datagram::EcatDatagram<TCommand, uint8_t> datagram{ {{slave_address, buffer_regs[MailboxesRegs::EMPTY]}}, sm_status };

        do {
            telegram.transfer(datagram);
        } while((datagram.get_wkc() < 0x0001) || ((sm_status & 0x08) != 0));
    }

    template<typename TypeT>
    void empty(datagram::EcatTelegram& telegram) {
        using TCommand = command::EcatCommand<TypeT, command::RD>;

        auto slave_address = buffer_slave_.get_slave().get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        uint16_t sm_status{0x0000};
        datagram::EcatDatagram<TCommand, uint16_t> datagram{ {{slave_address, buffer_regs[MailboxesRegs::AVAILABLE]}}, sm_status };

        do {
            telegram.transfer(datagram);
        } while(datagram.get_wkc() < 0x0001);
    }

    void increment_counter() {
        counter_++;

        if(counter_ > 7) {
            counter_ = 1;
        }
    }

    template<typename TypeT, typename... DataTypes>
    void send_data(datagram::EcatTelegram& telegram, uint16_t channel, uint16_t priority, ProtocolType type, DataTypes&... data) {
        using TCommand = command::EcatCommand<TypeT, command::WR>;

        EcatSlave& slave = buffer_slave_.get_slave();
        auto slave_address = slave.get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        MailboxHeader header{
            .length = CustomTuple<DataTypes...>::size,
            .address = 0, // slave.get_slave_address<command::FP>()
            .channel = channel,
            .priority = priority,
            .type = type,
            .cnt = static_cast<uint16_t>(counter_)
        };
        datagram::EcatDatagram<TCommand, MailboxHeader, DataTypes...> datagram{ {{slave_address, buffer_regs[MailboxesRegs::WRITE]}}, header, data... };
        uint16_t padding = buffer_slave_.get_buffer_properties_write().length - sizeof(MailboxHeader) - CustomTuple<DataTypes...>::size;

        datagram.set_padding(padding);

        do {
            telegram.transfer(datagram);
        } while(datagram.get_wkc() < 0x0001);

        increment_counter();
    }

    template<typename TypeT, typename... DataTypes>
    void receive_data(datagram::EcatTelegram& telegram, DataTypes&... data) {
        using TCommand = command::EcatCommand<TypeT, command::RD>;

        auto slave_address = buffer_slave_.get_slave().get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        MailboxHeader header;
        datagram::EcatDatagram<TCommand, MailboxHeader, DataTypes...> datagram{ {{slave_address, buffer_regs[MailboxesRegs::READ]}}, header, data... };
        uint16_t padding = buffer_slave_.get_buffer_properties_read().length - sizeof(MailboxHeader) - CustomTuple<DataTypes...>::size;

        datagram.set_padding(padding);

        do {
            telegram.transfer(datagram);
        } while(datagram.get_wkc() < 0x0001);

        //DebugP_log("header.length = %d\r\n", header.length);
        //DebugP_log("header.address = %d\r\n", header.address);
        //DebugP_log("header.channel = %d\r\n", header.channel);
        //DebugP_log("header.priority = %d\r\n", header.priority);
        //DebugP_log("header.type = %d\r\n", header.type);
        //DebugP_log("header.cnt = %d\r\n", header.cnt);
    }

    template<typename TypeT, typename... DataTypes>
    void receive(datagram::EcatTelegram& telegram, DataTypes&... data) {
        wait_available<TypeT>(telegram);
        receive_data<TypeT, DataTypes...>(telegram, data...);
    }

    template<typename TypeT, typename... DataTypes>
    void send(datagram::EcatTelegram& telegram, uint16_t channel, uint16_t priority, ProtocolType type, DataTypes&... data) {
        empty<TypeT>(telegram);
        wait_empty<TypeT>(telegram);
        send_data<TypeT, DataTypes...>(telegram, channel, priority, type, data...);
        wait_empty<TypeT>(telegram);
    }

    template<typename TypeT, typename... DataTypes>
    void sdo_read(datagram::EcatTelegram& telegram, uint16_t index, uint8_t subindex, DataTypes&... data) {
        CoEElements elements{
            .coe_header = {
                .number = 0x00,
                .service = static_cast<uint8_t>(Service::SDO_REQUEST) },
            .command_specifier = {
                .size = 1,
                .transfer_type = static_cast<uint8_t>(TransferType::NORMAL),
                .data_set_size = 0,
                .complete_access = 0,
                .command_spec = static_cast<uint8_t>(SDOReqCommandSpecifier::UPLOAD)},
            .index = index,
            .subindex = subindex };
        CompleteSize complete_size{0x00000000};

        send<TypeT, CoEElements, CompleteSize>(telegram, 0, 0, ProtocolType::CoE, elements, complete_size);
        receive<TypeT, CoEElements, CompleteSize, DataTypes...>(telegram, elements, complete_size, data...);

        if( (elements.coe_header.service != static_cast<uint8_t>(Service::SDO_RESPONSE)) ||
            (elements.command_specifier.command_spec != static_cast<uint8_t>(SDOReqCommandSpecifier::UPLOAD)) ) {
            DebugP_log("CoE error: = 0x%04x\r\n", complete_size.value); // 0x601004 - The object cannot be accessed via complete access
        }

        //DebugP_log("elements.coe_header.number = %d\r\n", elements.coe_header.number);
        //DebugP_log("elements.coe_header.service = %d\r\n", elements.coe_header.service);
        //DebugP_log("elements.command_specifier.size = %d\r\n", elements.command_specifier.size);
        //DebugP_log("elements.command_specifier.transfer_type = %d\r\n", elements.command_specifier.transfer_type);
        //DebugP_log("elements.command_specifier.data_set_size = %d\r\n", elements.command_specifier.data_set_size);
        //DebugP_log("elements.command_specifier.complete_access = %d\r\n", elements.command_specifier.complete_access);
        //DebugP_log("elements.command_specifier.command_spec = %d\r\n", elements.command_specifier.command_spec);
        //DebugP_log("elements.index = %d\r\n", elements.index);
        //DebugP_log("elements.subindex = %d\r\n", elements.subindex);
        //DebugP_log("complete_size = %d\r\n", complete_size);
    }

    template<typename TypeT>
    uint16_t pdo_map_read(datagram::EcatTelegram& telegram, uint16_t pdo_map_index) {
        uint16_t pdo_data_size{0x0000}; // Размер данных в битах !
        uint8_t pdo_block_count{0x00};

        sdo_read<TypeT, uint8_t>(telegram, pdo_map_index, 0x00, pdo_block_count);

        DebugP_log("pdo_block_count = 0x%01x\r\n", pdo_block_count);

        for(uint8_t pdo_map_subindex = 1; pdo_map_subindex < (pdo_block_count + 1); pdo_map_subindex++) {
            uint16_t pdo_block_index{0x0000};
            sdo_read<TypeT, uint16_t>(telegram, pdo_map_index, pdo_map_subindex, pdo_block_index);

            //DebugP_log("pdo_block_index = 0x02%x\r\n", pdo_block_index);

            uint8_t pdo_block_object_count{0x00};
            sdo_read<TypeT, uint8_t>(telegram, pdo_block_index, 0, pdo_block_object_count);

            DebugP_log("pdo_block_object_count = 0x%01x\r\n", pdo_block_object_count);

            for(uint8_t pdo_block_subindex = 1; pdo_block_subindex < (pdo_block_object_count + 1); pdo_block_subindex++) {
                PDODescriptor descriptor;
                sdo_read<TypeT, PDODescriptor>(telegram, pdo_block_index, pdo_block_subindex, descriptor);

                pdo_data_size += descriptor.size;

                DebugP_log("descriptor.size = 0x%01x\r\n", descriptor.size);
                //DebugP_log("descriptor.subindex = 0x%01x\r\n", descriptor.subindex);
                //DebugP_log("descriptor.index = 0x%02x\r\n", descriptor.index);
            }
        }

        //DebugP_log("pdo_data_size = %d\r\n", pdo_data_size);

        return pdo_data_size/8;
    }

private:
    ecat_buffer::EcatBufferSlave& buffer_slave_;

    uint8_t counter_{0x00};
};

class EthEcatSdoMailbox {
public:
    EthEcatSdoMailbox(ecat_buffer::EthEcatBuffer& ecat_buffer): ecat_buffer_{ecat_buffer} { }

    void init();
    void pdo_map_read(PDOMap& map);

private:
    ecat_buffer::EthEcatBuffer& ecat_buffer_;

    std::vector<EcatSdoMailboxSlave> sdo_mailbox_slaves_;
};

} // namespace ecat_sdo_mailbox

} // namespace free_rtos

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_SDO_MAILBOX_HPP_ */
